# generated by datamodel-codegen:
#   timestamp: 2024-11-20T13:58:32+00:00

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Annotated, Any, ClassVar, Dict, List, Literal, Optional, Union

import pandas as pd
from pydantic import AwareDatetime, Field, RootModel

from mostlyai.base import CustomBaseModel


class PermissionLevel(str, Enum):
    """
    The permission level of the user or group with respect to this entity
    - VIEW: The user can view and use the entity
    - EDIT: The user can edit and share the entity
    - ADMIN: The user can edit, share and delete the entity
    There is only a single user per entity with ADMIN rights, and that user is the owner of the entity.

    """

    view = "VIEW"
    edit = "EDIT"
    admin = "ADMIN"


class Credits(CustomBaseModel):
    current: Annotated[
        Optional[float],
        Field(None, description="The credit balance for the current time period"),
    ] = None
    limit: Annotated[
        Optional[float],
        Field(
            None,
            description="The credit limit for the current time period. If empty, then there is no limit.",
        ),
    ] = None
    period_start: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="periodStart",
            description="The UTC date and time when the current time period started",
        ),
    ] = None
    period_end: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="periodEnd",
            description="The UTC date and time when the current time period ends",
        ),
    ] = None


class ParallelTrainingJobs(CustomBaseModel):
    current: Annotated[
        Optional[int],
        Field(None, description="The number of currently running training jobs"),
    ] = None
    limit: Annotated[
        Optional[int],
        Field(
            None,
            description="The maximum number of running training jobs at any time. If empty, then there is no limit.",
        ),
    ] = None


class ParallelGenerationJobs(CustomBaseModel):
    current: Annotated[
        Optional[int],
        Field(None, description="The number of currently running generation jobs"),
    ] = None
    limit: Annotated[
        Optional[int],
        Field(
            None,
            description="The maximum number of running generation jobs at any time. If empty, then there is no limit.",
        ),
    ] = None


class UserUsage(CustomBaseModel):
    credits: Optional[Credits] = None
    parallel_training_jobs: Annotated[
        Optional[ParallelTrainingJobs], Field(None, alias="parallelTrainingJobs")
    ] = None
    parallel_generation_jobs: Annotated[
        Optional[ParallelGenerationJobs], Field(None, alias="parallelGenerationJobs")
    ] = None


class NotificationStatus(str, Enum):
    """
    The status of the notification
    """

    unread = "UNREAD"
    read = "READ"


class NotificationType(str, Enum):
    """
    The type of the notification
    """

    resource_ready = "RESOURCE_READY"
    resource_shared = "RESOURCE_SHARED"
    resource_liked = "RESOURCE_LIKED"
    resource_failed = "RESOURCE_FAILED"


class Metadata(CustomBaseModel):
    created_at: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="createdAt",
            description="The UTC date and time when the resource has been created.\n",
            examples=["2023‐09‐07T18:40:39Z"],
        ),
    ] = None
    owner_id: Annotated[
        Optional[str],
        Field(
            None,
            alias="ownerId",
            description="The unique identifier of the owner of the entity",
        ),
    ] = None
    owner_name: Annotated[
        Optional[str],
        Field(
            None, alias="ownerName", description="The name of the owner of the entity"
        ),
    ] = None
    current_user_permission_level: Annotated[
        Optional[PermissionLevel], Field(None, alias="currentUserPermissionLevel")
    ] = None
    current_user_like_status: Annotated[
        Optional[bool],
        Field(
            None,
            alias="currentUserLikeStatus",
            description="A boolean indicating whether the user has liked the entity or not",
        ),
    ] = None
    short_lived_file_token: Annotated[
        Optional[str],
        Field(
            None,
            alias="shortLivedFileToken",
            description="An auto-generated short-lived file token (`slft`) for accessing resource artefacts.\nThe token is always restricted to a single resource, only valid for 60 minutes, and \nonly accepted by API endpoints that allow to download single files.\n",
        ),
    ] = None


class PaginatedTotalCount(RootModel[int]):
    root: Annotated[
        int, Field(description="The total number of entities within the list")
    ]


class ModelEncodingType(str, Enum):
    """
    The encoding type used for model training and data generation.
    - AUTO: Model chooses among available encoding types based on the column's data type.
    - TABULAR_CATEGORICAL: Model samples from existing (non-rare) categories.
    - TABULAR_NUMERIC_AUTO: Model chooses among 3 numeric encoding types based on the values.
    - TABULAR_NUMERIC_DISCRETE: Model samples from existing discrete numerical values.
    - TABULAR_NUMERIC_BINNED: Model samples from binned buckets, to then sample randomly within a bucket.
    - TABULAR_NUMERIC_DIGIT: Model samples each digit of a numerical value.
    - TABULAR_CHARACTER: Model samples each character of a string value.
    - TABULAR_DATETIME: Model samples each part of a datetime value.
    - TABULAR_DATETIME_RELATIVE: Model samples the relative difference between datetimes within a sequence.
    - TABULAR_LAT_LONG: Model samples a latitude-longitude column. The format is "latitude,longitude".
    - LANGUAGE_TEXT: Model will train a distinct LANGUAGE model for this column, to then generate free text.
    # - LANGUAGE_NUMERIC
    # - LANGUAGE_CATEGORICAL
    # - LANGUAGE_DATETIME
    Encoding types, that are not being prefixed with either TABULAR or LANGUAGE have been deprecated.

    """

    auto = "AUTO"
    tabular_categorical = "TABULAR_CATEGORICAL"
    tabular_numeric_auto = "TABULAR_NUMERIC_AUTO"
    tabular_numeric_discrete = "TABULAR_NUMERIC_DISCRETE"
    tabular_numeric_binned = "TABULAR_NUMERIC_BINNED"
    tabular_numeric_digit = "TABULAR_NUMERIC_DIGIT"
    tabular_character = "TABULAR_CHARACTER"
    tabular_datetime = "TABULAR_DATETIME"
    tabular_datetime_relative = "TABULAR_DATETIME_RELATIVE"
    tabular_lat_long = "TABULAR_LAT_LONG"
    language_text = "LANGUAGE_TEXT"
    categorical = "CATEGORICAL"
    numeric_auto = "NUMERIC_AUTO"
    numeric_discrete = "NUMERIC_DISCRETE"
    numeric_binned = "NUMERIC_BINNED"
    numeric_digit = "NUMERIC_DIGIT"
    character = "CHARACTER"
    datetime = "DATETIME"
    datetime_relative = "DATETIME_RELATIVE"
    lat_long = "LAT_LONG"


class ConnectorAccessType(str, Enum):
    source = "SOURCE"
    destination = "DESTINATION"


class ConnectorType(str, Enum):
    mysql = "MYSQL"
    postgres = "POSTGRES"
    mssql = "MSSQL"
    oracle = "ORACLE"
    mariadb = "MARIADB"
    snowflake = "SNOWFLAKE"
    bigquery = "BIGQUERY"
    databricks = "DATABRICKS"
    azure_storage = "AZURE_STORAGE"
    google_cloud_storage = "GOOGLE_CLOUD_STORAGE"
    s3_storage = "S3_STORAGE"
    file_upload = "FILE_UPLOAD"
    hive = "HIVE"


class ConnectorTestConnection(RootModel[bool]):
    root: Annotated[
        bool,
        Field(
            description="If true, the connection will be tested before saving. In case of error, the connector will not be saved.\nIf false, the connection will not be tested.\n"
        ),
    ]


class ConnectorUsage(CustomBaseModel):
    no_of_shares: Annotated[
        Optional[int],
        Field(
            None, alias="noOfShares", description="Number of shares of this connector."
        ),
    ] = None
    no_of_generators: Annotated[
        Optional[int],
        Field(
            None,
            alias="noOfGenerators",
            description="Number of generators using this connector.",
        ),
    ] = None


class ConnectorListItem(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a connector")]
    name: Annotated[str, Field(description="The name of a connector")]
    type: ConnectorType
    access_type: Annotated[ConnectorAccessType, Field(alias="accessType")]
    metadata: Metadata
    usage: Optional[ConnectorUsage] = None


class Connector(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a connector")]
    name: Annotated[str, Field(description="The name of a connector")]
    type: ConnectorType
    access_type: Annotated[ConnectorAccessType, Field(alias="accessType")]
    config: Optional[Dict[str, Any]] = None
    secrets: Optional[Dict[str, str]] = None
    ssl: Optional[Dict[str, str]] = None
    metadata: Optional[Metadata] = None
    usage: Optional[ConnectorUsage] = None
    table_id: Annotated[
        Optional[str],
        Field(
            None,
            alias="tableId",
            description="Optional. ID of a source table or a synthetic table, that this connector belongs to.\nIf not set, then this connector is managed independently of any generator or synthetic dataset.\n",
        ),
    ] = None
    OPEN_URL_PARTS: ClassVar[list] = ["d", "connectors"]

    def update(self, config) -> "Connector":
        """
        Update a connector, and optionally validate the connection before saving.

        If validation fails, a 400 status with an error message will be returned.

        For the structure of the config, secrets and ssl parameters, see the CREATE method.

        :return: The updated connector
        """
        return self.client._update(connector_id=self.id, config=config)

    def delete(self):
        """
        Delete connector
        """
        return self.client._delete(connector_id=self.id)

    def locations(self, prefix: str = "") -> list:
        """
        List connector locations

        List the available databases, schemas, tables or folders for a connector.
        For storage connectors, this returns list of folders and files at root, respectively at `prefix` level.
        For DB connectors, this returns list of schemas (or databases for DBs without schema), respectively list of tables if `prefix` is provided.

        The formats of the locations are:

        - Cloud storage:
            - `AZURE_STORAGE`: `container/path`
            - `GOOGLE_CLOUD_STORAGE`: `bucket/path`
            - `S3_STORAGE`: `bucket/path`
        - Database:
            - `BIGQUERY`: `dataset.table`
            - `DATABRICKS`: `schema.table`
            - `HIVE`: `database.table`
            - `MARIADB`: `database.table`
            - `MSSQL`: `schema.table`
            - `MYSQL`: `database.table`
            - `ORACLE`: `schema.table`
            - `POSTGRES`: `schema.table`
            - `SNOWFLAKE`: `schema.table`
        :param prefix: The prefix to filter the results by.
        :return: A list of locations (schemas, databases, directories, etc.) on the given level.
        """
        return self.client._locations(connector_id=self.id, prefix=prefix)

    def schema(self, location: str) -> list[dict[str, Any]]:
        """
        Retrieve the schema (column names, original data types and default model encoding types) of the table at a connector location.

        Please refer to `locations()` for the format of the location.

        :param location: The location of the table
        :return: The retrieved schema
        """
        return self.client._schema(connector_id=self.id, location=location)

    def config(self) -> dict[str, Any]:
        """
        Retrieve writeable generator properties

        :return: The generator properties as dictionary
        """
        return self.client._config(connector_id=self.id)

    def shares(self):
        return self.client._shares(resource=self)


class GeneratorUsage(CustomBaseModel):
    total_datapoints: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalDatapoints",
            description="The total number of datapoints generated by this generator.",
        ),
    ] = None
    total_compute_time: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalComputeTime",
            description="The total compute time in seconds used for training this generator.\nThis is the sum of the compute time of all trained tasks.\n",
        ),
    ] = None
    no_of_synthetic_datasets: Annotated[
        Optional[int],
        Field(
            None,
            alias="noOfSyntheticDatasets",
            description="Number of synthetic datasets generated by this generator.",
        ),
    ] = None
    no_of_shares: Annotated[
        Optional[int],
        Field(
            None, alias="noOfShares", description="Number of shares of this generator."
        ),
    ] = None
    no_of_likes: Annotated[
        Optional[int],
        Field(
            None, alias="noOfLikes", description="Number of likes of this generator."
        ),
    ] = None


class SourceTableData(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="The base64-encoded string derived from a Parquet file containing the specified source table.\n"
        ),
    ]


class SourceTableIncludeChildren(RootModel[bool]):
    root: Annotated[
        bool,
        Field(
            description="If true, all tables that are referenced by foreign keys will\nbe included. If false, only the selected table will be included.\n"
        ),
    ]


class SourceColumnValueRange(CustomBaseModel):
    """
    The (privacy-safe) range of values detected within a source column. These values can then be used as seed values
    for conditional generation. For CATEGORICAL and NUMERIC_DISCRETE encoding types, this will be given as a list
    of unique values, sorted by popularity. For other NUMERIC and for DATETIME encoding types, this will be given
    as a min and max value. Note, that this property will only be populated, once the analysis step for the training
    of the generator has been completed.

    """

    min: Annotated[
        Optional[str],
        Field(
            None,
            description="The minimum value of the column. For dates, this is represented in ISO format.",
        ),
    ] = None
    max: Annotated[
        Optional[str],
        Field(
            None,
            description="The maximum value of the column. For dates, this is represented in ISO format.",
        ),
    ] = None
    values: Annotated[
        Optional[List[str]],
        Field(
            None,
            description="The list of distinct values of the column. Limited to a maximum of 1000 values.",
        ),
    ] = None
    has_null: Annotated[
        Optional[bool],
        Field(None, description="If true, null value was detected within the column."),
    ] = None


class SourceColumn(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a source column")]
    name: Annotated[str, Field(description="The name of a source column")]
    included: Annotated[
        bool,
        Field(
            description="If true, the column will be included in the training.\nIf false, the column will be excluded from the training.\n"
        ),
    ]
    model_encoding_type: Annotated[ModelEncodingType, Field(alias="modelEncodingType")]
    value_range: Annotated[
        Optional[SourceColumnValueRange], Field(None, alias="valueRange")
    ] = None


class SourceForeignKey(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a foreign key")]
    column: Annotated[
        Optional[str], Field(None, description="The column name of a foreign key.")
    ] = None
    referenced_table: Annotated[
        str,
        Field(
            alias="referencedTable",
            description="The table name of the referenced table. That table must have a primary key already defined.",
        ),
    ]
    is_context: Annotated[
        bool,
        Field(
            alias="isContext",
            description="If true, then the foreign key will be considered as a context relation.\nNote, that only one foreign key relation per table can be a context relation.\n",
        ),
    ]


class ModelType(str, Enum):
    tabular = "TABULAR"
    language = "LANGUAGE"


class RareCategoryReplacementMethod(str, Enum):
    """
    Specifies, if the rare categories for categoricals will be replaced by a constant
    _RARE_ or by a sample from non-rare categories.
    Only applicable if valueProtection is set to True.

    """

    constant = "CONSTANT"
    sample = "SAMPLE"


class ModelConfigurationDifferentialPrivacy(CustomBaseModel):
    """
    The optional differential privacy configuration for training the model.
    If not provided, then no differential privacy will be applied.

    """

    max_epsilon: Annotated[
        Optional[float],
        Field(
            None,
            alias="maxEpsilon",
            description="Specifies the maximum allowable epsilon value. If the training process exceeds this threshold, it will be terminated early. Only model checkpoints with epsilon values below this limit will be retained. \nIf not provided, the training will proceed without early termination based on epsilon constraints.\n",
            ge=0.0,
            le=10000.0,
        ),
    ] = None
    noise_multiplier: Annotated[
        Optional[float],
        Field(
            1.5,
            alias="noiseMultiplier",
            description="The ratio of the standard deviation of the Gaussian noise to the L2-sensitivity of the function to which the noise is added (How much noise to add).\n",
            ge=0.0,
            le=10000.0,
        ),
    ] = 1.5
    max_grad_norm: Annotated[
        Optional[float],
        Field(
            1.0,
            alias="maxGradNorm",
            description="The maximum norm of the per-sample gradients for training the model with differential privacy.\n",
            ge=0.0,
            le=10000.0,
        ),
    ] = 1.0


class Accuracy(CustomBaseModel):
    overall: Annotated[
        Optional[float],
        Field(
            None,
            description="The overall accuracy of the model.\nThis is the average of univariate, bivariate and coherence accuracies.\nAccuracy is defined as (100% - Total Variation Distance) for each distribution, and then averaged across.\n",
            ge=0.0,
            le=1.0,
        ),
    ] = None
    univariate: Annotated[
        Optional[float],
        Field(
            None,
            description="The average accuracy of the discretized univariate distributions.\n",
            ge=0.0,
            le=1.0,
        ),
    ] = None
    bivariate: Annotated[
        Optional[float],
        Field(
            None,
            description="The average accuracy of the discretized bivariate distributions.\n",
            ge=0.0,
            le=1.0,
        ),
    ] = None
    coherence: Annotated[
        Optional[float],
        Field(
            None,
            description="The average accuracy of the discretized coherence distributions.\nOnly applicable for sequential models.\n",
            ge=0.0,
            le=1.0,
        ),
    ] = None
    overall_max: Annotated[
        Optional[float],
        Field(
            None,
            alias="overallMax",
            description="The expected accuracy of a holdout dataset of same size as synthetic dataset.\nThis serves as a reference for the overall accuracy of the trained generator.\n",
            ge=0.0,
            le=1.0,
        ),
    ] = None


class Distances(CustomBaseModel):
    dcr_training: Annotated[
        Optional[float],
        Field(
            None,
            alias="dcrTraining",
            description="The average distance to closest record (L2) between synthetic and training samples within the embedding space.",
            ge=0.0,
        ),
    ] = None
    dcr_holdout: Annotated[
        Optional[float],
        Field(
            None,
            alias="dcrHoldout",
            description="The average distance to closest record (L2) between synthetic and holdout samples within the embedding space.",
            ge=0.0,
        ),
    ] = None
    dcr_original: Annotated[
        Optional[float],
        Field(
            None,
            alias="dcrOriginal",
            description="(DEPRECATED) The average DCR between the original records.",
            ge=0.0,
        ),
    ] = None
    dcr_synthetic: Annotated[
        Optional[float],
        Field(
            None,
            alias="dcrSynthetic",
            description="(DEPRECATED) The average DCR between the synthetic and the original records.",
            ge=0.0,
        ),
    ] = None


class Similarity(CustomBaseModel):
    cosine_similarity_training_synthetic: Annotated[
        Optional[float],
        Field(
            None,
            alias="cosineSimilarityTrainingSynthetic",
            description="The cosine similarity between the training samples centroid and the synthetic samples centroid within the embedding space.",
            ge=-1.0,
            le=1.0,
        ),
    ] = None
    cosine_similarity_training_holdout: Annotated[
        Optional[float],
        Field(
            None,
            alias="cosineSimilarityTrainingHoldout",
            description="The cosine similarity between the training samples centroid and the holdout samples centroid within the embedding space.",
            ge=-1.0,
            le=1.0,
        ),
    ] = None
    discriminator_auc_training_synthetic: Annotated[
        Optional[float],
        Field(
            None,
            alias="discriminatorAUCTrainingSynthetic",
            description="The cross-validated average AUC of discriminating between training embeddings ans synthetic embeddings.",
            ge=0.0,
            le=1.0,
        ),
    ] = None
    discriminator_auc_training_holdout: Annotated[
        Optional[float],
        Field(
            None,
            alias="discriminatorAUCTrainingHoldout",
            description="The cross-validated average AUC of discriminating between training embeddings ans holdout embeddings.",
            ge=0.0,
            le=1.0,
        ),
    ] = None


class ModelMetrics(CustomBaseModel):
    accuracy: Optional[Accuracy] = None
    distances: Optional[Distances] = None
    similarity: Optional[Similarity] = None


class StepCode(str, Enum):
    pull_training_data = "PULL_TRAINING_DATA"
    analyze_training_data = "ANALYZE_TRAINING_DATA"
    encode_training_data = "ENCODE_TRAINING_DATA"
    train_model = "TRAIN_MODEL"
    generate_model_report_data = "GENERATE_MODEL_REPORT_DATA"
    create_model_report = "CREATE_MODEL_REPORT"
    finalize_training = "FINALIZE_TRAINING"
    pull_context_data = "PULL_CONTEXT_DATA"
    generate_data = "GENERATE_DATA"
    create_data_report = "CREATE_DATA_REPORT"
    finalize_generation = "FINALIZE_GENERATION"
    deliver_data = "DELIVER_DATA"


class ProgressValue(CustomBaseModel):
    value: Optional[int] = None
    max: Optional[int] = None


class ProgressStatus(str, Enum):
    """
    The status of a job or a step.
    NEW: The job/step is being configured, and has not started yet
    CONTINUE: The job/step is being configured, but has existing artefacts
    ON_HOLD: The job/step has been started, but is kept on hold
    QUEUED: The job/step has been started, and is awaiting for resources to execute
    IN_PROGRESS: The job/step is currently running
    DONE: The job/step has finished successfully
    FAILED: The job/step has failed
    CANCELED: The job/step has been canceled

    """

    new = "NEW"
    continue_ = "CONTINUE"
    on_hold = "ON_HOLD"
    queued = "QUEUED"
    in_progress = "IN_PROGRESS"
    done = "DONE"
    failed = "FAILED"
    canceled = "CANCELED"


class SyntheticDatasetUsage(CustomBaseModel):
    total_datapoints: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalDatapoints",
            description="The number of datapoints in the synthetic dataset",
        ),
    ] = None
    total_credits: Annotated[
        Optional[float],
        Field(
            None,
            alias="totalCredits",
            description="The number of credits used for the synthetic dataset",
        ),
    ] = None
    total_compute_time: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalComputeTime",
            description="The total compute time in seconds used for generating this synthetic dataset.\nThis is the sum of the compute time of all trained tasks.\n",
        ),
    ] = None
    no_of_shares: Annotated[
        Optional[int],
        Field(
            None,
            alias="noOfShares",
            description="Number of shares of this synthetic dataset.",
        ),
    ] = None
    no_of_likes: Annotated[
        Optional[int],
        Field(
            None,
            alias="noOfLikes",
            description="Number of likes of this synthetic dataset.",
        ),
    ] = None


class SyntheticDatasetListItem(CustomBaseModel):
    id: Annotated[
        str, Field(description="The unique identifier of a synthetic dataset")
    ]
    metadata: Metadata
    name: Annotated[str, Field(description="The name of a synthetic dataset")]
    description: Annotated[
        Optional[str], Field(None, description="The description of a synthetic dataset")
    ] = None
    generation_status: Annotated[ProgressStatus, Field(alias="generationStatus")]
    generation_time: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="generationTime",
            description="The UTC date and time when the generation has finished.",
        ),
    ] = None
    usage: Optional[SyntheticDatasetUsage] = None


class SyntheticDatasetFormat(str, Enum):
    csv = "CSV"
    parquet = "PARQUET"
    xlsx = "XLSX"


class SyntheticDatasetReportType(str, Enum):
    model = "MODEL"
    data = "DATA"


class SyntheticTableRebalancing(CustomBaseModel):
    """
    Configure rebalancing of the table.
    Only applicable for categorical columns of a subject table.

    """

    column: Annotated[
        Optional[str],
        Field(
            None,
            description="The name of the column to be rebalanced.\nThat column must be of modelEncodingType CATEGORICAL.\n",
        ),
    ] = None
    probabilities: Annotated[
        Optional[Dict[str, float]],
        Field(
            None,
            description="The target distribution of samples values.\nThe keys are the categorical values, and the values are the probabilities.\n",
            examples=[[{"US": 0.8}, {"male": 0.5, "female": 0.5}]],
        ),
    ] = None


class SyntheticTableFairness(CustomBaseModel):
    """
    Configure a fairness objective for the table. Only applicable for a subject table.
    The generated synthetic data will maintain robust statistical parity between the target column and
    the specified sensitive columns. All these columns must be categorical.

    """

    target_column: Annotated[str, Field(alias="targetColumn")]
    sensitive_columns: Annotated[List[str], Field(alias="sensitiveColumns")]


class SyntheticTableName(RootModel[str]):
    root: Annotated[
        str,
        Field(
            description="The name of a synthetic table. This matches the name of a corresponding SourceTable."
        ),
    ]


class ForeignKey(CustomBaseModel):
    column: Annotated[str, Field(description="The column name of a foreign key.")]
    referenced_table: Annotated[
        str,
        Field(
            alias="referencedTable",
            description="The table name of the referenced table. That table must have a primary key already defined.",
        ),
    ]
    is_context: Annotated[
        bool,
        Field(
            alias="isContext",
            description="If true, then the foreign key will be considered as a context relation.\nNote, that only one foreign key relation per table can be a context relation.\n",
        ),
    ]


class SyntheticDatasetDelivery(CustomBaseModel):
    overwrite_tables: Annotated[
        bool,
        Field(
            alias="overwriteTables",
            description="If true, tables in the destination will be overwritten.\nIf false, any tables exist, the delivery will fail.\n",
        ),
    ]
    destination_connector_id: Annotated[
        str,
        Field(
            alias="destinationConnectorId",
            description="The unique identifier of a connector",
        ),
    ]
    location: Annotated[
        str, Field(description="The location for the destination connector.")
    ]


class AssistantLiteLlmExtraItem(CustomBaseModel):
    key: Optional[str] = None
    value: Optional[str] = None


class LiteLlm(CustomBaseModel):
    model: Annotated[
        Optional[str],
        Field(
            None,
            description="The LiteLLM model of the assistant. See https://docs.litellm.ai/docs/providers.",
            examples=[["openai/gpt-3.5-turbo", "mistral/mistral-tiny"]],
        ),
    ] = None
    api_key: Annotated[
        Optional[str],
        Field(
            None,
            alias="apiKey",
            description="The API key for the selected LiteLLM model. See https://docs.litellm.ai/docs/providers.",
        ),
    ] = None
    extra: Annotated[
        Optional[List[AssistantLiteLlmExtraItem]],
        Field(
            None,
            description="Any additional configuration parameters for the selected LiteLLM model. See https://docs.litellm.ai/docs/providers.",
        ),
    ] = None


class DataLlm(CustomBaseModel):
    api_key: Annotated[
        Optional[str],
        Field(
            None,
            alias="apiKey",
            description="The API key for the DataLLM service. See https://data.mostly.ai.",
        ),
    ] = None


class AssistantSettings(CustomBaseModel):
    """
    Additional optional assistant settings used for LiteLLM
    """

    is_enabled: Annotated[
        Optional[bool],
        Field(
            None, alias="isEnabled", description="If true, the assistant is enabled."
        ),
    ] = None
    lite_llm: Annotated[Optional[LiteLlm], Field(None, alias="liteLlm")] = None
    data_llm: Annotated[Optional[DataLlm], Field(None, alias="dataLlm")] = None
    system_instructions: Annotated[
        Optional[str],
        Field(
            None,
            alias="systemInstructions",
            description="The system instructions of the assistant",
        ),
    ] = None
    custom_instructions: Annotated[
        Optional[str],
        Field(
            None,
            alias="customInstructions",
            description="The custom instructions of the assistant",
        ),
    ] = None
    default_system_instructions: Annotated[
        Optional[str],
        Field(
            None,
            alias="defaultSystemInstructions",
            description="The system instructions of the assistant",
        ),
    ] = None


class AssistantThreadSessionStatus(str, Enum):
    initializing = "initializing"
    running = "running"
    expired = "expired"


class AssistantTokenUsage(RootModel[float]):
    root: float


class AssistantMessageRole(str, Enum):
    """
    The role of the author of this message
    """

    system = "system"
    user = "user"
    assistant = "assistant"
    tool = "tool"
    heartbeat = "heartbeat"


class AssistantMessageContentType(str, Enum):
    """
    The type of the message content
    """

    text = "text"
    python = "python"
    console = "console"
    file = "file"


class AssistantMessageFinishReason(str, Enum):
    stop = "stop"
    length = "length"
    tool_calls = "tool_calls"


class AssistantMessage(CustomBaseModel):
    """
    A complete message.
    """

    id: Annotated[
        Optional[str],
        Field(None, description="The unique identifier of a assistant message"),
    ] = None
    role: Optional[AssistantMessageRole] = None
    content_type: Annotated[
        Optional[AssistantMessageContentType], Field(None, alias="contentType")
    ] = None
    content: Annotated[
        Optional[str],
        Field(
            None, description="The content of a message", max_length=65000, min_length=0
        ),
    ] = None
    finish_reason: Annotated[
        Optional[AssistantMessageFinishReason], Field(None, alias="finishReason")
    ] = None
    tokens_consumed: Annotated[
        Optional[float],
        Field(
            None,
            alias="tokensConsumed",
            description="The number of tokens consumed by the assistant message.",
        ),
    ] = None


class AssistantMessageDelta(CustomBaseModel):
    """
    A partial message delta generated by streamed model responses.
    """

    id: Annotated[
        Optional[str],
        Field(None, description="The unique identifier of a assistant message"),
    ] = None
    role: Optional[AssistantMessageRole] = None
    content_type: Annotated[
        Optional[AssistantMessageContentType], Field(None, alias="contentType")
    ] = None
    delta: Annotated[
        Optional[str],
        Field(
            None,
            description="The partial content of a message",
            max_length=65000,
            min_length=1,
        ),
    ] = None
    finish_reason: Annotated[
        Optional[AssistantMessageFinishReason], Field(None, alias="finishReason")
    ] = None
    tokens_consumed: Annotated[
        Optional[float],
        Field(
            None,
            alias="tokensConsumed",
            description="The number of tokens consumed by the model.",
        ),
    ] = None


class AssistantThreadUsage(CustomBaseModel):
    no_of_shares: Annotated[
        Optional[int],
        Field(
            None,
            alias="noOfShares",
            description="Number of shares of this assistant thread.",
        ),
    ] = None
    total_tokens_consumed: Annotated[
        Optional[float],
        Field(
            None,
            alias="totalTokensConsumed",
            description="The total number of tokens consumed by the thread.",
        ),
    ] = None


class ComputeResources(CustomBaseModel):
    cpus: Annotated[
        Optional[int], Field(None, description="The number of CPU cores")
    ] = None
    memory: Annotated[
        Optional[float], Field(None, description="The amount of memory in GB")
    ] = None
    gpus: Annotated[Optional[int], Field(0, description="The number of GPUs")] = 0
    gpu_memory: Annotated[
        Optional[float],
        Field(0, alias="gpuMemory", description="The amount of GPU memory in GB"),
    ] = 0


class ComputeListItem(CustomBaseModel):
    id: Optional[str] = None
    type: Optional[Literal["KUBERNETES"]] = None
    name: Optional[str] = None
    resources: Optional[ComputeResources] = None


class Compute(CustomBaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    type: Optional[Literal["KUBERNETES"]] = None
    config: Optional[Dict[str, Any]] = None
    secrets: Optional[Dict[str, Any]] = None
    resources: Optional[ComputeResources] = None
    order_index: Annotated[
        Optional[int],
        Field(
            None,
            alias="orderIndex",
            description="The index for determining the sort order when listing computes",
        ),
    ] = None


class UsageReportPeriod(CustomBaseModel):
    """
    The usage report for a specific month
    """

    period_start: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="periodStart",
            description="The UTC date and time when the reported time period started",
        ),
    ] = None
    period_end: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="periodEnd",
            description="The UTC date and time when the reported time period started",
        ),
    ] = None
    total_datapoints: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalDatapoints",
            description="The number of datapoints generated during the reported period",
        ),
    ] = None
    total_rows: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalRows",
            description="The number of rows generated during the reported period",
        ),
    ] = None
    total_credits: Annotated[
        Optional[float],
        Field(
            None,
            alias="totalCredits",
            description="The number of credits used during the reported period",
        ),
    ] = None


class BaseResource(CustomBaseModel):
    id: Annotated[
        Optional[str], Field(None, description="The unique identifier of the entity")
    ] = None
    name: Annotated[
        Optional[str], Field(None, description="The name of the entity")
    ] = None
    uri: Annotated[
        Optional[str],
        Field(
            None,
            description="The API service endpoint of the entity",
            examples=["/generators/94c77249-42bf-443a-8e17-6e18a19d60b8"],
        ),
    ] = None
    current_user_permission_level: Annotated[
        Optional[PermissionLevel], Field(None, alias="currentUserPermissionLevel")
    ] = None
    current_user_like_status: Annotated[
        Optional[bool],
        Field(
            None,
            alias="currentUserLikeStatus",
            description="A boolean indicating whether the user has liked the entity or not",
        ),
    ] = None


class User(CustomBaseModel):
    id: Annotated[
        Optional[str], Field(None, description="The unique identifier of a user")
    ] = None
    first_name: Annotated[
        Optional[str],
        Field(None, alias="firstName", description="First name of a user"),
    ] = None
    last_name: Annotated[
        Optional[str], Field(None, alias="lastName", description="Last name of a user")
    ] = None
    email: Annotated[Optional[str], Field(None, description="The email of a user")] = (
        None
    )


class CurrentUser(CustomBaseModel):
    id: Annotated[
        Optional[str], Field(None, description="The unique identifier of a user")
    ] = None
    first_name: Annotated[
        Optional[str],
        Field(None, alias="firstName", description="First name of a user"),
    ] = None
    last_name: Annotated[
        Optional[str], Field(None, alias="lastName", description="Last name of a user")
    ] = None
    email: Annotated[Optional[str], Field(None, description="The email of a user")] = (
        None
    )
    settings: Optional[Dict[str, Any]] = None
    usage: Optional[UserUsage] = None
    unread_notifications: Annotated[
        Optional[int],
        Field(
            None,
            alias="unreadNotifications",
            description="Number of unread notifications for the user",
        ),
    ] = None


class Notification(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of the notification")]
    type: NotificationType
    message: Annotated[str, Field(description="The message of the notification")]
    status: NotificationStatus
    created_at: Annotated[
        AwareDatetime,
        Field(
            alias="createdAt",
            description="The UTC date and time when the notification has been created",
        ),
    ]
    resource_uri: Annotated[
        Optional[str],
        Field(
            None,
            alias="resourceUri",
            description="The API service endpoint of the entity",
            examples=["/generators/94c77249-42bf-443a-8e17-6e18a19d60b8"],
        ),
    ] = None


class GeneratorListItem(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a generator")]
    name: Annotated[
        Optional[str], Field(None, description="The name of a generator")
    ] = None
    description: Annotated[
        Optional[str], Field(None, description="The description of a generator")
    ] = None
    training_status: Annotated[ProgressStatus, Field(alias="trainingStatus")]
    training_time: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="trainingTime",
            description="The UTC date and time when the training has finished.",
        ),
    ] = None
    usage: Optional[GeneratorUsage] = None
    metadata: Metadata


class Probe(CustomBaseModel):
    name: Annotated[Optional[str], Field(None, description="The name of the table")] = (
        None
    )
    rows: Optional[List[Dict[str, Any]]] = None


class ModelConfiguration(CustomBaseModel):
    """
    The training configuration for a AI model
    """

    model: Annotated[
        Optional[str],
        Field(
            None,
            description="The model to be used for training.",
            examples=[
                [
                    "MOSTLY_AI/Small",
                    "MOSTLY_AI/Medium",
                    "MOSTLY_AI/Large",
                    "MOSTLY_AI/LSTMFromScratch-3m",
                    "microsoft/phi-1_5",
                ]
            ],
        ),
    ] = None
    max_sample_size: Annotated[
        Optional[int],
        Field(
            None,
            alias="maxSampleSize",
            description="The maximum number of samples to consider for training.\nIf not provided, then all available samples will be taken.\n",
            ge=1,
            le=1000000000,
        ),
    ] = None
    batch_size: Annotated[
        Optional[int],
        Field(
            None,
            alias="batchSize",
            description="The batch size used for training the model.\nIf not provided, batchSize will be chosen automatically.\n",
            ge=1,
            le=1000000,
        ),
    ] = None
    max_training_time: Annotated[
        Optional[int],
        Field(
            10,
            alias="maxTrainingTime",
            description="The maximum number of minutes to train the model.",
            ge=0,
            le=100000,
        ),
    ] = 10
    max_epochs: Annotated[
        Optional[int],
        Field(
            100,
            alias="maxEpochs",
            description="The maximum number of epochs to train the model.",
            ge=0,
            le=100000,
        ),
    ] = 100
    max_sequence_window: Annotated[
        Optional[int],
        Field(
            100,
            alias="maxSequenceWindow",
            description="The maximum sequence window to consider for training.\nOnly applicable for TABULAR models.\n",
            ge=1,
            le=100000,
        ),
    ] = 100
    enable_flexible_generation: Annotated[
        Optional[bool],
        Field(
            True,
            alias="enableFlexibleGeneration",
            description="If true, then the trained generator can be used for rebalancing and imputation.\nOnly applicable for TABULAR models.\n",
        ),
    ] = True
    value_protection: Annotated[
        Optional[bool],
        Field(
            True,
            alias="valueProtection",
            description="Defines if Rare Category, Extreme value, or Sequence length protection will be applied.\n",
        ),
    ] = True
    rare_category_replacement_method: Annotated[
        Optional[RareCategoryReplacementMethod],
        Field(
            "CONSTANT",
            alias="rareCategoryReplacementMethod",
            description="Specifies, if the rare categories for categoricals will be replaced by a constant\n_RARE_ or by a sample from non-rare categories.\nOnly applicable if valueProtection is set to True.\n",
        ),
    ] = "CONSTANT"
    differential_privacy: Annotated[
        Optional[ModelConfigurationDifferentialPrivacy],
        Field(None, alias="differentialPrivacy"),
    ] = None
    compute: Optional[str] = None


class ProgressStep(CustomBaseModel):
    id: Optional[str] = None
    model_label: Annotated[
        Optional[str],
        Field(
            None,
            alias="modelLabel",
            description="The unique label for the model, consisting of table name and a suffix for the model type.\nThis will be empty for steps that are not related to a model.\n",
            examples=[["census:tabular", "census:language"]],
        ),
    ] = None
    compute_name: Annotated[Optional[str], Field(None, alias="computeName")] = None
    restarts: Annotated[
        Optional[int],
        Field(
            None, description="Number of previous restarts for the corresponding task."
        ),
    ] = None
    step_code: Annotated[Optional[StepCode], Field(None, alias="stepCode")] = None
    start_date: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="startDate",
            description="The UTC date and time when the job has started.\nIf the job has not started yet, then this is None.\n",
            examples=["2024-01-25T12:34:56Z"],
        ),
    ] = None
    end_date: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="endDate",
            description="The UTC date and time when the job has ended.\nIf the job is still, then this is None.\n",
            examples=["2024-01-25T12:34:56Z"],
        ),
    ] = None
    messages: Optional[List[Dict[str, Any]]] = None
    error_message: Annotated[Optional[str], Field(None, alias="errorMessage")] = None
    progress: Optional[ProgressValue] = None
    status: Optional[ProgressStatus] = None


class SyntheticTableConfiguration(CustomBaseModel):
    """
    The sample configuration for a synthetic table
    """

    sample_size: Annotated[
        Optional[int],
        Field(
            None,
            alias="sampleSize",
            description="Number of generated samples. Only applicable for subject tables.\nIf neither size nor seed is provided, then the default behavior for Synthetic Datasets is to generate the\nsame size of samples as the original, and the default behavior for Synthetic Datasets is to generate one\nsubject only.\n",
            ge=1,
        ),
    ] = None
    sample_seed_connector_id: Annotated[
        Optional[str],
        Field(
            None,
            alias="sampleSeedConnectorId",
            description="The connector id of the seed data for conditional generation.\nOnly applicable for subject tables.\n",
        ),
    ] = None
    sample_seed_dict: Annotated[
        Optional[str],
        Field(
            None,
            alias="sampleSeedDict",
            description="The base64-encoded string derived from a json line file containing the specified sample seed data.\n",
        ),
    ] = None
    sample_seed_data: Annotated[
        Optional[str],
        Field(
            None,
            alias="sampleSeedData",
            description="The base64-encoded string derived from a Parquet file containing the specified sample seed data.\n",
        ),
    ] = None
    sampling_temperature: Annotated[
        Optional[float],
        Field(
            None,
            alias="samplingTemperature",
            description="temperature for sampling",
            ge=0.0,
            le=10.0,
        ),
    ] = None
    sampling_top_p: Annotated[
        Optional[float],
        Field(
            None, alias="samplingTopP", description="topP for sampling", ge=0.0, le=1.0
        ),
    ] = None
    rebalancing: Optional[SyntheticTableRebalancing] = None
    imputation: Annotated[
        Optional[List[str]],
        Field(
            None,
            description="Specify a list of column names that are to be imputed.\nImputed columns will suppress the sampling of NULL values.\n",
        ),
    ] = None
    fairness: Optional[SyntheticTableFairness] = None
    tabular_compute: Annotated[Optional[str], Field(None, alias="tabularCompute")] = (
        None
    )
    language_compute: Annotated[Optional[str], Field(None, alias="languageCompute")] = (
        None
    )


class SyntheticTable(CustomBaseModel):
    id: Annotated[
        Optional[str],
        Field(None, description="The unique identifier of a synthetic table"),
    ] = None
    name: Annotated[
        Optional[str],
        Field(
            None,
            description="The name of a source table. It must be unique within a generator.",
        ),
    ] = None
    configuration: Optional[SyntheticTableConfiguration] = None
    model_metrics: Annotated[
        Optional[ModelMetrics], Field(None, alias="modelMetrics")
    ] = None
    language_model_metrics: Annotated[
        Optional[ModelMetrics], Field(None, alias="languageModelMetrics")
    ] = None
    foreign_keys: Annotated[
        Optional[List[ForeignKey]],
        Field(None, alias="foreignKeys", description="The foreign keys of this table."),
    ] = None
    total_rows: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalRows",
            description="The total number of rows for that table in the generated synthetic dataset.\n",
        ),
    ] = None
    total_datapoints: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalDatapoints",
            description="The total number of datapoints for that table in the generated synthetic dataset.\n",
        ),
    ] = None
    source_table_total_rows: Annotated[
        Optional[int],
        Field(
            None,
            alias="sourceTableTotalRows",
            description="The total number of rows in the source table while fetching data for training.\n",
        ),
    ] = None


class AssistantThreadListItem(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a assistant thread")]
    metadata: Metadata
    name: Annotated[
        Optional[str], Field(None, description="The name of a assistant thread")
    ] = None
    usage: Optional[AssistantThreadUsage] = None


class AssistantThread(CustomBaseModel):
    id: Annotated[
        Optional[str],
        Field(None, description="The unique identifier of a assistant thread"),
    ] = None
    metadata: Optional[Metadata] = None
    name: Annotated[
        Optional[str], Field(None, description="The name of a assistant thread")
    ] = None
    session_status: Annotated[
        Optional[AssistantThreadSessionStatus], Field(None, alias="sessionStatus")
    ] = None
    messages: Annotated[
        Optional[List[AssistantMessage]],
        Field(
            None,
            description="List of all existing messages, excluding any system message.\nExample:\n  What's the square root of 9?   role: user        type: text\n  ok, I will write some python   role: assistant   type: text\n  r = math.sqrt(9)\\nr            role: assistant   type: python\n  3                              role: tool        type: console\n  The answer is 3!               role: assistant   type: text\n  Plot me a random barplot       role: user        type: text\n  ... plt.savefig() ...          role: assistant   type: python\n  Here is ![img]() ...           role: assistant   type: text\n  Write me a random file         role: user        type: text\n  ... .to_csv() ...              role: assistant   type: python\n  Here is [file]() ...           role: assistant   type: text\n",
        ),
    ] = None
    usage: Optional[AssistantThreadUsage] = None


class Share(User):
    permission_level: Annotated[
        Optional[PermissionLevel], Field(None, alias="permissionLevel")
    ] = None


class SourceTable(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a source table")]
    source_connector: Annotated[
        Optional[BaseResource], Field(None, alias="sourceConnector")
    ] = None
    location: Annotated[
        Optional[str],
        Field(
            None,
            description="The location of a source table. Together with the source connector it uniquely\nidentifies a source, and samples data from there.\n",
        ),
    ] = None
    name: Annotated[
        str,
        Field(
            description="The name of a source table. It must be unique within a generator."
        ),
    ]
    primary_key: Annotated[
        Optional[str],
        Field(
            None, alias="primaryKey", description="The column name of the primary key"
        ),
    ] = None
    columns: Annotated[
        List[SourceColumn], Field(description="The columns of this generator table.")
    ]
    foreign_keys: Annotated[
        Optional[List[SourceForeignKey]],
        Field(None, alias="foreignKeys", description="The foreign keys of a table."),
    ] = None
    model_metrics: Annotated[
        Optional[ModelMetrics], Field(None, alias="modelMetrics")
    ] = None
    language_model_metrics: Annotated[
        Optional[ModelMetrics], Field(None, alias="languageModelMetrics")
    ] = None
    model_configuration: Annotated[
        Optional[ModelConfiguration], Field(None, alias="modelConfiguration")
    ] = None
    language_model_configuration: Annotated[
        Optional[ModelConfiguration], Field(None, alias="languageModelConfiguration")
    ] = None
    total_rows: Annotated[
        Optional[int],
        Field(
            None,
            alias="totalRows",
            description="The total number of rows in the source table while fetching data for training.\n",
        ),
    ] = None

    def model_qa_report(self):
        if self.client and hasattr(self.client, "model_qa_report"):
            return self.client.model_qa_report(
                generator_id=self.extra_key_values["generator_id"], table_id=self.id
            )

    def model_samples(self, **kwargs):
        if self.client and hasattr(self.client, "model_samples"):
            return self.client.model_qa_report(
                generator_id=self.extra_key_values["generator_id"],
                table_id=self.id,
                **kwargs,
            )

    def get_column(self, column_id: str):
        if self.client and hasattr(self.client, "get_column"):
            return self.client.get_column(
                generator_id=self.extra_key_values["generator_id"],
                table_id=self.id,
                column_id=column_id,
            )

    def create_foreign_key(self, **kwargs):
        if self.client and hasattr(self.client, "create_foreign_key"):
            return self.client.create_foreign_key(
                generator_id=self.extra_key_values["generator_id"],
                table_id=self.id,
                **kwargs,
            )

    def update_foreign_key(self, **kwargs):
        if self.client and hasattr(self.client, "update_foreign_key"):
            return self.client.update_foreign_key(
                generator_id=self.extra_key_values["generator_id"],
                table_id=self.id,
                **kwargs,
            )

    def delete_foreign_key(self, **kwargs):
        if self.client and hasattr(self.client, "delete_foreign_key"):
            return self.client.delete_foreign_key(
                generator_id=self.extra_key_values["generator_id"],
                table_id=self.id,
                **kwargs,
            )


class JobProgress(CustomBaseModel):
    id: Optional[str] = None
    start_date: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="startDate",
            description="The UTC date and time when the job has started.\nIf the job has not started yet, then this is None.\n",
            examples=["2024-01-25T12:34:56Z"],
        ),
    ] = None
    end_date: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="endDate",
            description="The UTC date and time when the job has ended.\nIf the job is still, then this is None.\n",
            examples=["2024-01-25T12:34:56Z"],
        ),
    ] = None
    progress: Optional[ProgressValue] = None
    status: Optional[ProgressStatus] = None
    steps: Optional[List[ProgressStep]] = None


class SyntheticDataset(CustomBaseModel):
    id: Annotated[
        str, Field(description="The unique identifier of a synthetic dataset")
    ]
    generator: Optional[BaseResource] = None
    metadata: Metadata
    name: Annotated[str, Field(description="The name of a synthetic dataset")]
    description: Annotated[
        Optional[str], Field(None, description="The description of a synthetic dataset")
    ] = None
    generation_status: Annotated[ProgressStatus, Field(alias="generationStatus")]
    generation_time: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="generationTime",
            description="The UTC date and time when the generation has finished.",
        ),
    ] = None
    tables: Annotated[
        Optional[List[SyntheticTable]],
        Field(None, description="The tables of this synthetic dataset."),
    ] = None
    delivery: Optional[SyntheticDatasetDelivery] = None
    accuracy: Annotated[
        Optional[float],
        Field(
            None,
            description="The overall accuracy of the trained generator.\nThis is the average of the overall accuracy scores of all trained models.\n",
        ),
    ] = None
    usage: Optional[SyntheticDatasetUsage] = None
    OPEN_URL_PARTS: ClassVar[list] = ["d", "synthetic-datasets"]
    generation: Annotated[Optional[Any], Field(exclude=True)] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generation = self.Generation(self)

    def update(self, config) -> "SyntheticDataset":
        """
        Update synthetic dataset

        See config for the structure of the parameters.

        :return: The updated synthetic dataset
        """
        return self.client._update(synthetic_dataset_id=self.id, config=config)

    def delete(self):
        """
        Delete synthetic dataset
        """
        return self.client._delete(synthetic_dataset_id=self.id)

    def config(self) -> dict[str, Any]:
        """
        Retrieve writeable synthetic dataset properties

        :return: The synthetic dataset properties as dictionary
        """
        return self.client._config(synthetic_dataset_id=self.id)

    def download(
        self,
        format: SyntheticDatasetFormat = "PARQUET",
        file_path: Union[str, Path, None] = None,
    ) -> Path:
        """
        Download synthetic dataset and save to file

        :param format: The format of the synthetic dataset
        :param file_path: The file path to save the synthetic dataset
        """
        bytes, filename = self.client._download(
            synthetic_dataset_id=self.id,
            ds_format=format,
            short_lived_file_token=self.metadata.short_lived_file_token,
        )
        file_path = Path(file_path or ".")
        if file_path.is_dir():
            file_path = file_path / filename
        file_path.write_bytes(bytes)
        return file_path

    def data(
        self, return_type: Literal["auto", "dict"] = "auto"
    ) -> Union[pd.DataFrame, dict[str, pd.DataFrame]]:
        """
        Download synthetic dataset and return as dictionary of pandas DataFrames

        :return: The synthetic dataset as dictionary of pandas DataFrames
        """
        dfs = self.client._data(
            synthetic_dataset_id=self.id,
            short_lived_file_token=self.metadata.short_lived_file_token,
        )
        if return_type == "auto" and len(dfs) == 1:
            return list(dfs.values())[0]
        else:
            return dfs

    def shares(self):
        return self.client._shares(resource=self)

    class Generation:
        def __init__(self, _synthetic_dataset: "SyntheticDataset"):
            self.synthetic_dataset = _synthetic_dataset

        def start(self) -> None:
            """
            Start generation
            """
            return self.synthetic_dataset.client._generation_start(
                self.synthetic_dataset.id
            )

        def cancel(self) -> None:
            """
            Cancel generation
            """
            return self.synthetic_dataset.client._generation_cancel(
                self.synthetic_dataset.id
            )

        def progress(self) -> JobProgress:
            """
            Retrieve job progress of generation
            """
            return self.synthetic_dataset.client._generation_progress(
                self.synthetic_dataset.id
            )

        def wait(self, progress_bar: bool, interval: float = 2) -> "SyntheticDataset":
            """
            Poll generation progress and loop until generation has completed

            :param progress_bar: If true, then the progress bar will be displayed.
            :param interval: The interval in seconds to poll the job progress
            """
            return self.synthetic_dataset.client._generation_wait(
                self.synthetic_dataset.id, progress_bar=progress_bar, interval=interval
            )


class Generator(CustomBaseModel):
    id: Annotated[str, Field(description="The unique identifier of a generator")]
    name: Annotated[
        Optional[str], Field(None, description="The name of a generator")
    ] = None
    description: Annotated[
        Optional[str], Field(None, description="The description of a generator")
    ] = None
    training_status: Annotated[ProgressStatus, Field(alias="trainingStatus")]
    training_time: Annotated[
        Optional[AwareDatetime],
        Field(
            None,
            alias="trainingTime",
            description="The UTC date and time when the training has finished.",
        ),
    ] = None
    usage: Optional[GeneratorUsage] = None
    metadata: Metadata
    accuracy: Annotated[
        Optional[float],
        Field(
            None,
            description="The overall accuracy of the trained generator.\nThis is the average of the overall accuracy scores of all trained models.\n",
        ),
    ] = None
    tables: Annotated[
        Optional[List[SourceTable]],
        Field(None, description="The tables of this generator"),
    ] = None
    OPEN_URL_PARTS: ClassVar[list] = ["d", "generators"]
    training: Annotated[Optional[Any], Field(exclude=True)] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.training = self.Training(self)

    def update(self, config) -> "Generator":
        """
        Update generator

        See config for the structure of the parameters.

        :return: The updated generator
        """
        return self.client._update(generator_id=self.id, config=config)

    def delete(self):
        """
        Delete generator
        """
        return self.client._delete(generator_id=self.id)

    def config(self) -> dict[str, Any]:
        """
        Retrieve writeable generator properties

        :return: The generator properties as dictionary
        """
        return self.client._config(generator_id=self.id)

    def shares(self):
        return self.client._shares(resource=self)

    def export_to_file(
        self,
        file_path: Union[str, Path, None] = None,
    ) -> Path:
        """
        Export generator and save to file

        :param file_path: The file path to save the synthetic dataset
        """
        bytes, filename = self.client._export_to_file(generator_id=self.id)
        file_path = Path(file_path or ".")
        if file_path.is_dir():
            file_path = file_path / filename
        file_path.write_bytes(bytes)
        return file_path

    def clone(self, training_status: Literal["NEW", "CONTINUE"] = "NEW") -> "Generator":
        """
        Clone generator

        :param training_status: The training status of the cloned generator
        """
        return self.client._clone(generator_id=self.id, training_status=training_status)

    class Training:
        def __init__(self, _generator: "Generator"):
            self.generator = _generator

        def start(self) -> None:
            """
            Start training
            """
            return self.generator.client._training_start(self.generator.id)

        def cancel(self) -> None:
            """
            Cancel training
            """
            return self.generator.client._training_cancel(self.generator.id)

        def progress(self) -> JobProgress:
            """
            Retrieve job progress of training
            """
            return self.generator.client._training_progress(self.generator.id)

        def wait(self, progress_bar: bool, interval: float = 2) -> "Generator":
            """
            Poll training progress and loop until training has completed

            :param progress_bar: If true, then the progress bar will be displayed.
            :param interval: The interval in seconds to poll the job progress
            """
            return self.generator.client._training_wait(
                self.generator.id, progress_bar=progress_bar, interval=interval
            )

        def list_synthetic_dataset(self) -> list["SyntheticDataset"]:
            """
            List synthetic datasets

            List the synthetic datasets that were created based on this generator.

            :return: A list of synthetic datasets
            """
            raise "Not implemented yet."
            # return self.generator.client._list_synthetic_datasets(self.generator.id)
            pass
